/* (C)opyright MMIV-MMV Anselm R. Garbe <garbeam at gmail dot com>
* 
* Changes: 
*    Robert Manea, 10/18/2006 
*            Write some text to see slock is running
*            echo password as '*' to screen
*
* See LICENSE file for license details.
*/
#define _XOPEN_SOURCE 500

#if HAVE_SHADOW_H
#include <shadow.h>
#else
#include <pwd.h>
#endif

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <X11/keysym.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

Display *dpy;
Window w;
XFontStruct *font_info;
int screen;
GC gc;

int
clear_pwd (int s, int e) {
    XSetForeground(dpy, gc, BlackPixel(dpy, screen));
    for(; s!=e; --e) {
        XDrawString(dpy, w, gc, 
                50 + e * (XTextWidth(font_info, "*", 1) + 2), 
                DisplayHeight(dpy, screen) - (font_info->ascent + font_info->descent) - 50, 
                "*", 1); 
    } 
    XSetForeground(dpy, gc, WhitePixel(dpy, screen));

    return e;
}

int
main(int argc, char **argv) {
    char curs[] = {0, 0, 0, 0, 0, 0, 0, 0};
    char buf[32], passwd[256];
    char text[256]="screen locked";
    int num, i;
#if HAVE_SHADOW_H
    struct spwd *sp;
#else
    struct passwd *pw;
#endif
    unsigned int len;
    Bool running = True;
    Cursor invisible;
    KeySym ksym;
    Pixmap pmap;
    XColor black, dummy;
    XEvent ev;
    XSetWindowAttributes wa;
    XGCValues gv;
    char *dtext = text, *font  = "*-helvetica-*-24-*";
    if((argc > 1) && !strncmp(argv[1], "-v", 3)) {
        fputs("slock-"VERSION", (C)opyright MMVI Anselm R. Garbe\n", stdout);
        exit(EXIT_SUCCESS);
    }
    if((argc > 1) && !strncmp(argv[1], "-t", 3)) {
	strcpy(text, argv[2]);
    }
    if(geteuid() != 0) {
        fputs("slock: cannot retrieve password entry (make sure to suid slock)\n", stderr);
        exit(EXIT_FAILURE);
    }
#if HAVE_SHADOW_H
    sp = getspnam(getenv("USER"));
    endspent();
#else
    pw = getpwuid(getuid());
    endpwent();
#endif
    if(!(dpy = XOpenDisplay(0))) {
        fputs("slock: cannot open display\n", stderr);
        exit(EXIT_FAILURE);
    }
    screen = DefaultScreen(dpy);

    /* init */
    passwd[0] = 0;

    wa.override_redirect = 1;
    wa.background_pixel = BlackPixel(dpy, screen);
    w = XCreateWindow(dpy, RootWindow(dpy, screen), 0, 0,
            DisplayWidth(dpy, screen), DisplayHeight(dpy, screen),
            0, DefaultDepth(dpy, screen), CopyFromParent,
            DefaultVisual(dpy, screen), CWOverrideRedirect | CWBackPixel, &wa);

    if ( (gc = XCreateGC(dpy, w, 0, &gv)) < 0) {
        fputs("slock: cannot initialize GC\n", stderr);
        exit(EXIT_FAILURE);
    }

    if( !(font_info = XLoadQueryFont(dpy, font))) {
        fputs("slock: cannot load specified font\n", stderr);
        exit(EXIT_FAILURE);
    }
        
    XSetFont(dpy, gc, font_info->fid);
    XSetForeground(dpy, gc, WhitePixel(dpy, screen));
    XSetBackground(dpy, gc, BlackPixel(dpy, screen));

    XAllocNamedColor(dpy, DefaultColormap(dpy, screen), "black", &black, &dummy);
    pmap = XCreateBitmapFromData(dpy, w, curs, 8, 8);
    invisible = XCreatePixmapCursor(dpy, pmap, pmap, &black, &black, 0, 0);
    XDefineCursor(dpy, w, invisible);
    running = XGrabPointer(dpy, RootWindow(dpy, screen), False,
            ButtonPressMask | ButtonReleaseMask | PointerMotionMask,
            GrabModeAsync, GrabModeSync, None, invisible, CurrentTime) == GrabSuccess
        && XGrabKeyboard(dpy, RootWindow(dpy, screen), True, GrabModeAsync,
                GrabModeAsync, CurrentTime) == GrabSuccess;


    XMapRaised(dpy, w);
    XSync(dpy, False);


    XDrawString( dpy, w, gc, 
            (DisplayWidth (dpy, screen) - XTextWidth(font_info, dtext, strlen(dtext))) / 2, 
            (DisplayHeight(dpy, screen) - (font_info->ascent + font_info->descent)) / 2, 
            dtext, strlen(dtext) );




    /* main event loop */
    while(running && !XNextEvent(dpy, &ev)) {
    //    XNextEvent(dpy, &ev);
        if(ev.type == KeyPress) {
            len = strlen(passwd);
            buf[0] = 0;
            num = XLookupString(&ev.xkey, buf, sizeof(buf), &ksym, 0);
            if(IsFunctionKey(ksym) || IsKeypadKey(ksym)
                    || IsMiscFunctionKey(ksym) || IsPFKey(ksym)
                    || IsPrivateKeypadKey(ksym))
                continue;
            switch(ksym) {
                case XK_Return:
#if HAVE_SHADOW_H
                    if((running = strncmp(crypt(passwd, sp->sp_pwdp), sp->sp_pwdp, sizeof(passwd)))) 
#else
                        if((running = strncmp(crypt(passwd, pw->pw_passwd), pw->pw_passwd, sizeof(passwd)))) 
#endif
                            XBell(dpy, 100);
                    passwd[0] = 0;
                    i = clear_pwd(0, i);
                    break;
                case XK_Escape:
                    passwd[0] = 0;
                    i = clear_pwd(0, i);
                    break;
                case XK_BackSpace:
                    if(len) 
                        passwd[--len] = 0;
                    if(i)
                        i = clear_pwd(i-1, i);
                    break;
                default:
                    if(num && !iscntrl((int) buf[0])) {
                        buf[num] = 0;
                        if(len)
                            strncat(passwd, buf, sizeof(passwd));
                        else
                            strncpy(passwd, buf, sizeof(passwd));

                        ++i;
                        XDrawString(dpy, w, gc, 
                                50 + i * (XTextWidth(font_info, "*", 1) + 2), 
                                DisplayHeight(dpy, screen) - (font_info->ascent + font_info->descent) - 50, 
                                "*", 1); 
                    }
                    break;
            }
        }
    }



    XUngrabPointer(dpy, CurrentTime);
    XFreePixmap(dpy, pmap);
    XDestroyWindow(dpy, w);
    XCloseDisplay(dpy);
    return 0;
}
